#ifndef FILLBRANCH_H
#define FILLBRANCH_H

#include <fstream>
#include <stdio.h>
#include <TFile.h>
#include <TTree.h>
#include <TH1D.h>
#include <string>
#include <yaml-cpp/yaml.h>
#include <iostream>
//#include "CreateBranch.h" 
using namespace std;

#define MaxWords 200
#define MaxSegs 16
#define MaxParas 9 // the maximum paramters of each Seg, for example, the SegId 1 has two branches, fRf and fF1ppac, then the number of parameters is 2
#define MaxDims 9 // the maximun dimensions of each parameters
#define MaxChans 96 //the maximum channels for each dimension
class FillBranch{
   private:
       TFile * rootfile;
       TTree *tree;
       TH1D* scaler; /* 10/5 by Shimizu  */
   public:
      int NumOfPara[MaxSegs]; //number of paramters for each SegId
      int NumOfCh[MaxSegs][MaxParas];//number of channels for each parameter
      int NumOfDim[MaxSegs][MaxParas];//number of dimenstions for each parameter
      short Bran_1d[MaxSegs][MaxParas][MaxChans]; 
      short Bran_nd[MaxSegs][MaxParas][MaxDims][MaxChans];
      short SegData[MaxSegs][MaxWords]; 
      int SegFlag[MaxSegs];
      int MaxSegId;
      int SegId_F1ppac;
      int ParaId_F1ppac;
      int SegId_Ppac;
      int FirstId[MaxSegs][MaxParas];
      string DecodeType[MaxSegs];
      string MapName[MaxSegs];

      void ReadRecipe(Int_t RunNum);
      void CreateBranch();
      void CreateRootFile();
      void BranchFill(Int_t SegId, short Bran[MaxSegs][MaxWords]);
      void clear(){
       memset(Bran_1d,-10,sizeof(Bran_1d));
       memset(Bran_nd,-10,sizeof(Bran_nd));
       for(int i=0;i<MaxSegs;i++){
           for(int j=0;j<MaxWords;j++){
               SegData[i][j]=-10;
           }
       }
      };
      void FillTree(){tree->Fill();};
      void RecodeScaler(int *data); /* 10/5 by Shimizu */
      void Print(int SegId);
};

//=============================================================================

void FillBranch::ReadRecipe(Int_t RunNum){
   char rdfName[100], RootFileName[100];
   sprintf(rdfName,"./rdf/%04d.rdf",RunNum);
   sprintf(RootFileName,"./rootfile/%04d.root",RunNum);
   YAML::Node recipe = YAML::LoadFile("./recipe.yaml"); // Recipe file 
   MaxSegId = 0;
   for(int i=0;i<MaxSegs;i++){SegFlag[i]=0;}
   try{ // read parameters by yaml-cpp
    string infile = rdfName;
    string outfile = RootFileName;

    string trname = recipe["tree name"].as<string>();
    cout << infile << endl;
    cout << outfile << endl;
    cout << trname << endl;
   rootfile=new TFile(outfile.c_str(),"recreate");
   tree=new TTree(trname.c_str(),"experiment tree");
    YAML::Node setup = recipe["setup"].as<YAML::Node>();    
    for(unsigned i=0;i<setup.size();i++){
      int count = 0;
      Int_t iseg = setup[i]["segment no."].as<int>();
      if(iseg > MaxSegId) MaxSegId = iseg;
      string dtype = setup[i]["dtype"].as<string>();
      string mapfile = setup[i]["mapfile"].as<string>();
      DecodeType[iseg] = dtype;
    //  sprintf(MapName[iseg],"./map/%s",mapfile);
       //  MapName[iseg] = "./map_hayakawa/" + mapfile;
      MapName[iseg] = "./cordfer_map/" + mapfile;
      SegFlag[iseg]=1;

      YAML::Node bra = setup[i]["branches"].as<YAML::Node>();      
      for(unsigned j=0;j<bra.size();j++){
	Int_t Ndim = bra[j]["name"][1].as<int>();
	Int_t Nch = bra[j]["name"][2].as<int>();
	FirstId[iseg][j] = bra[j]["name"][4].as<int>();
        NumOfCh[iseg][j] = Nch; 
        NumOfDim[iseg][j] = Ndim; 
        count++;
      } // for j
      NumOfPara[iseg] = count;
    } // for i   
  }catch(YAML::Exception& e){ // For error display
    cerr << e.what() << endl;
  }
}
//=============================================================================

void FillBranch::CreateBranch(){
 
   YAML::Node recipe = YAML::LoadFile("./recipe.yaml"); // Recipe file 
  try{ // read parameters by yaml-cpp
    // Input and output files

    YAML::Node setup = recipe["setup"].as<YAML::Node>();    
    for(unsigned i=0;i<setup.size();i++){
      Int_t iseg = setup[i]["segment no."].as<int>();
      YAML::Node bra = setup[i]["branches"].as<YAML::Node>();      
      for(unsigned j=0;j<bra.size();j++){
        string bname = bra[j]["name"][0].as<string>();
	Int_t Ndim = bra[j]["name"][1].as<int>();
//	Int_t Nch = bra[j]["name"][2].as<int>();
	string datatype = bra[j]["name"][3].as<string>();
        if(bname=="fF1ppac"){SegId_F1ppac=iseg;ParaId_F1ppac=j;}
        else if(bname=="fPpac"){SegId_Ppac=iseg;}
        Char_t brancharray[100];
       if(Ndim==1){
       sprintf(brancharray,"%s[%d]/S",bname.c_str(),MaxChans);
       tree->Branch(bname.c_str(),Bran_1d[iseg][j],brancharray);    
       }
       else if(Ndim>=2){
       sprintf(brancharray,"%s[%d][%d]/S",bname.c_str(),Ndim,MaxChans);
       tree->Branch(bname.c_str(),Bran_nd[iseg][j],brancharray);    
       }
     } // for j
    } // for i    
  }catch(YAML::Exception& e){ // For error display
    cerr << e.what() << endl;
  }
}

//=============================================================================

void FillBranch::CreateRootFile(){
         rootfile->Write();
         rootfile->Close();
      }

//=============================================================================

void FillBranch::BranchFill(Int_t SegId, short SegData[][MaxWords]){
      cout<<dec;
      int para = NumOfPara[SegId];//number of branches for this SegId
      int dim[para]; //dimentions of each branch array
      int ch[para]; //number of channels for each dimenstion
      for(int i=0;i<para;i++){
         dim[i] = NumOfDim[SegId][i];
         ch[i]  = NumOfCh[SegId][i];
      }
//      cout<<"SegId: "<<SegId<<endl; 
//      cout<<"numb of parameters: "<<para<<endl;
//      cout<<"parameterID"<<"  "<<"dimentions"<<"  "<<"channels"<<"  "<<"First Id"<<endl;
//      for(int i=0;i<para;i++){
//        cout<<i<<"    "<<dim[i]<<"    "<<ch[i]<<"  "<<FirstId[SegId][i]<<endl;
//      } 

   for(int i=0;i<para;i++){
        if(DecodeType[SegId]=="CAEN"||DecodeType[SegId]=="LeCroy"){
           if(dim[i]==1){
            for(int j=0;j<ch[i];j++){
             Bran_1d[SegId][i][j] = SegData[SegId][FirstId[SegId][i]+j-1];
            }
           }
          else if(dim[i]>1){
           for(int j=0;j<dim[i];j++){
             for(int k=0;k<ch[i];k++){
                Bran_nd[SegId][i][j][k] = SegData[SegId][FirstId[SegId][i]+k+j*ch[i]-1];
             }
           }
         }
        }// end of CAEN or LeCroy

       else{
       if(dim[i]==1){
           for(int j=0;j<ch[i];j++){
           Bran_1d[SegId][i][j] = SegData[SegId][FirstId[SegId][i]+j];
          }
         }
        else if(dim[i]>1){
         for(int j=0;j<dim[i];j++){
            for(int k=0;k<ch[i];k++){
                Bran_nd[SegId][i][j][k] = SegData[SegId][FirstId[SegId][i]+k+j*ch[i]];
            }
         }
       }
 //    StartCh += dim[i]*ch[i];
       }
   }//end of loop of para
}

//=============================================================================

void FillBranch::Print(int SegId){
      int para = NumOfPara[SegId];//number of branches for this SegId
      int dim[para]; //dimentions of each branch array
      int ch[para]; //number of channels for each dimenstion
      for(int i=0;i<para;i++){
         dim[i] = NumOfDim[SegId][i];
         ch[i]  = NumOfCh[SegId][i];
      }
      cout<<dec;
      cout<<"@@@@@@@@@@@@ output of SegId: "<<SegId<<"  @@@@@@@@@@@@@"<<endl;
    
     for(int i=0;i<para;i++){
         cout<<"para Id:"<< i<<endl;
         if(dim[i]==1){
             for(int j=0;j<ch[i];j++){
                 cout<<Bran_1d[SegId][i][j]<<"  ";
             }
             cout<<endl;
         }
         if(dim[i]>1){
             for(int j=0;j<dim[i];j++){
                 for(int k=0;k<ch[i];k++){
                     cout<<Bran_nd[SegId][i][j][k]<<"  ";
                 }
                 cout<<endl;
             }
             cout<<endl;
         }
     }
     cout<<"@@@@@@@@@@@@ END of output of SegId: "<<SegId<<"  @@@@@@@@@@@@@"<<endl;
     cout<<endl;
}

//=============================================================================

void FillBranch::RecodeScaler(int *data){ /* 10/5 by Shimizu */

    scaler = new TH1D("Scaler", "" , 16, 1, 16);
    scaler->GetXaxis()->SetBinLabel(1,"Ungated");
    scaler->GetXaxis()->SetBinLabel(2,"Gated");
    scaler->GetXaxis()->SetBinLabel(3,"Beam*SSD");
    scaler->GetXaxis()->SetBinLabel(4,"Beam/n");
    scaler->GetXaxis()->SetBinLabel(5,"F3PPACa");
    scaler->GetXaxis()->SetBinLabel(6,"F3PPACb");
    scaler->GetXaxis()->SetBinLabel(7,"PSD-OR");
    scaler->GetXaxis()->SetBinLabel(8,"Clock 10Hz");
    scaler->GetXaxis()->SetBinLabel(9,"BRam");
    scaler->GetXaxis()->SetBinLabel(10,"N/A");
    scaler->GetXaxis()->SetBinLabel(11,"N/A");
    scaler->GetXaxis()->SetBinLabel(12,"N/A");
    scaler->GetXaxis()->SetBinLabel(13,"N/A");
    scaler->GetXaxis()->SetBinLabel(14,"N/A");
    scaler->GetXaxis()->SetBinLabel(15,"N/A");
    scaler->GetXaxis()->SetBinLabel(16,"N/A");

    scaler->Fill(1,data[0]);
    scaler->Fill(2,data[1]);
    scaler->Fill(3,data[2]);
    scaler->Fill(4,data[3]);
    scaler->Fill(5,data[4]);
    scaler->Fill(6,data[5]);
    scaler->Fill(7,data[6]);
    scaler->Fill(8,data[7]);
    scaler->Fill(9,data[8]);
    scaler->Fill(10,data[9]);
    scaler->Fill(11,data[10]);
    scaler->Fill(12,data[11]);
    scaler->Fill(13,data[12]);
    scaler->Fill(14,data[13]);
    scaler->Fill(15,data[14]);
    scaler->Fill(16,data[15]);

}
#endif






