#include "DataDecoder.h"
//#include "ChannelMap.h"

using namespace std;


DataDecoder::DataDecoder() {
  SetDecoderType("Raw");
}

DataDecoder::DataDecoder(const char *dtypename) {
  SetDecoderType(dtypename);
}
  
DataDecoder::DataDecoder(const char *dtypename,const char *mapfilename) {
  SetDecoderType(dtypename);
  ReadMap(mapfilename);
}


DataDecoder::~DataDecoder() {

}

int DataDecoder::SetDecoderType(const char *dtypename) {
  switch (dtypename[0]) {
    case 'L' :
    case 'l' :
      decoder_type=3;
      strcpy(decoder_type_name,"LeCroy");
      break;

    case 'C' :
    case 'c' :
      decoder_type=4;
      strcpy(decoder_type_name,"CAEN");

      break;
     case 'B' :
     case 'b' :
       decoder_type=2;
       strcpy(decoder_type_name,"Bit");
       break;
     default :
       decoder_type=1;
       strcpy(decoder_type_name,"Raw");
       break;
 } 

return 0;
}

//short *decoded...data array must be prepared by the caller.
int DataDecoder::Decode(int nwords, short *raw, short *decoded) {
  int len=0;
  switch (decoder_type) {
  case 1: len=DecodeRaw(nwords,raw,decoded);
    break;
  case 2: len=DecodeBit(nwords,raw,decoded);
    break;
  case 3: len=DecodeLeCroy(nwords,raw,decoded);
    break;
  case 4: len=DecodeCaen(nwords,raw,decoded);
    break;
  default:
    cerr << "Invalid decoder type" << endl;
    break;

  }


  return len;

}
int DataDecoder::DecodeRaw(int nwords, short *raw, short *decoded) {
  for (int i=0;i<nwords;i++) {
    decoded[i]=raw[i];
  }
  return nwords;
}

int DataDecoder::DecodeBit(int nwords, short *raw, short *decoded) {

  int nbits = sizeof (*raw) *8;
  for (int i=0;i<nwords;i++) {
    for (int j=0;j<nbits;j++) { //assumes 16-bit
      decoded[16*i+j]= ((raw[i] & (1 << j))!=0); 
    }
  }
  //  for (UInt_t i = 0; i < wordsN * 16; i++)
  //  (*reg)[i] = (TDataItem<Bool_t>(i + start, (data[0] & (1 << i))));
  return nwords*nbits;
}

int DataDecoder::DecodeLeCroy(int nwords, short *raw, short *decoded) {
  //Warning:
  //map file...id, module number channel number are always > 0.
  //channel number can be 0 in the raw data, so 0-7 are shifted to 1-8.
  //decoded[0] ...id= 1 datum will be stored.
 
  int id,maxid_data=0;

  if (!MapRead()) { //Map file was read?
    cerr << "Error: Map file is not read" << endl;
    return 1;
  }

  for (id = 0; id <= MaxId(); id++) { //clear array
    decoded[id]=0;
  }

  for (int i = 0; i < nwords; i++) {
 //   cout << "i=" <<  dec <<i << ", raw[i]=" << hex << raw[i] << endl;

    short header = raw[i];

    if ((header & 0x8000) == 0x8000) {
      short vsn = header & 0x00ff;

      if (vsn>12) {
	cerr << "DecodeLeCroy: Strange vsn=" << vsn << endl;
	vsn%=16; //in case of error in the header
      } 

      i++;      
      // silena adc vsn's
      short vdc = header >> 8 & 0x000f; //data count
      short pw  = raw[i++] & 0x00ff;
 
      for (int j = 0; j < vdc && i + j < nwords; j++) { //loop on one module data
	short sub = (raw[i + j] >> 12) & 0x0007;
	sub++; //if the map file starts from ch=1

	short val = raw[i + j] & 0x0fff;
	short ovf = (raw[i + j] >> 15) & 0x0001;
	 printf("vsn %.4x, vdc %.4x, pw %.4x, sub %.4x, val %.4x,"
	        " ovf %.4x\n", vsn, vdc, pw, sub, val, ovf);
	if (ovf) {
	  // cerr << "DecodeFera: WARNING - ADC overflow" << endl;
	  val=5000;
	} 
	/* ignore pattern word
	  else if (pw & 1 << sub) {
	  sub += (vsn - 1) * 8 + start;
	  module->push_back(TDataItem<UShort_t>(sub, val));
	} else {
	  cerr << "DecodeFera: ERROR - subaddress and pattern-word mismatch"
	       << endl;
	}
	*/
       id=Map(vsn,sub);
       //       cout << "id" << id << " vsn" <<vsn << " sub"<< sub<<endl;

       if (id>0) {
	 decoded[id-1] = val;
	 if (id>maxid_data) {maxid_data=id;}
       }
      } //loop on j

      i += vdc-1;
    }
    
    // unknown vsn's - shouldn't ever get here
    //    cerr << "DecodeFera: Unknown VSN = " << vsn << endl;
    //break;
  }


  return maxid_data;
}

int DataDecoder::DecodeCaen(int nwords, short *raw, short *decoded) {

  // V785 ADC is 12-bit
  int i = 0;
  int type = 0xffffffff;

  int *raw_i=(int *) raw; //change pointer type
  int nwords_i=nwords * sizeof(short)/sizeof(int);
  int id,maxid_data=0;

  struct Datum_t {
    short geo;
    short type;
    short channel;
    short valid;
    short underthreshold;
    short overflow;
    short value;
  } datum;

  //  struct Invalid_t {
  //  UShort_t type;
  //} invalid;

  if (!MapRead()) {
    cerr << "Error: Map file is not read" << endl;
    return 1;
  }


  for (id = 0; id <= MaxId(); id++) { //clear array
    decoded[id]=0;
  }


  for (i = 0; i < nwords_i; i++) {

    //    cout << "# "<< i << "/" << nwords_i << ": " << hex << raw_i[i] << dec << endl;
    //  return 0;

    type = (raw_i[i] & 0x07000000) >> 24;
    //cout << hex << raw_i[i] <<endl;

    switch (type) {
    case 2: //header, do nothing
      break;
    case 0: //normal data
      // valid datum
      datum.geo = (raw_i[i] & 0xf8000000) >> 27;
      // datum.type = (data[i] & 0x07000000) >> 24;
      datum.channel = (raw_i[i] & 0x001f0000) >> 16;

      datum.channel++; //if the map file starts from ch=1


      // datum.valid = (data[i] & 0x00002000) >> 14;
      // datum.underthreshold = (data[i] & 0x00001000) >> 13;
      datum.overflow = (raw_i[i] & 0x000008000) >> 12;
      datum.value = raw_i[i] & 0x00000fff;
      /*
       cout << "datum.geo = "            << datum.geo            << endl
            << "datum.channel = "        << datum.channel        << endl
            << "datum.type = "           << datum.type           << endl
            << "datum.valid = "          << datum.valid          << endl
            << "datum.underthreshold = " << datum.underthreshold << endl
            << "datum.overflow = "       << datum.overflow       << endl
            << "datum.value = "          << datum.value          << endl;
      */
      if (datum.overflow) {
	datum.value = 5000; //Don't know if it's good treatment or not
      }



       id=Map(datum.geo,datum.channel);
       //       cout << "XXX" <<id << "(" <<datum.geo << ", "<< datum.channel << ")" <<endl;
       if (id>0) {
	 decoded[id-1] = datum.value;
	 if (id>maxid_data) {maxid_data=id;}
       }
      //      module->push_back(TDataItem<UShort_t>((datum.geo - 1) * 32 + 
      //datum.channel + start,
      //                                            datum.value));

      break;
    case 4: //ender
      // end-of-block
      // eob.geo     = (data[i] & 0xf8000000) >> 27;
      // eob.type    = (data[i] & 0x07000000) >> 24;
      // eob.counter = data[i] & 0x00ffffff;
      // cout << "eob.geo = "     << eob.geo     << endl
      //      << "eob.type = "    << eob.type    << endl
      //      << "eob.counter = " << eob.counter << endl;
      //      cerr << "ender" << endl;
      break;
    case 6:
      // invalid datum
      //cerr << "DecodeCaen: ERROR - invalid word" << endl;
      break;
    default:
      cerr << "DecodeCaen: ERROR - invalid word" << endl;
      break;
    }
  }

  return maxid_data;
}

/*
void DataDecoder::DecodeRegister(const UShort_t *data, UInt_t wordsN, UInt_t start,
                              vector<TDataItem<Bool_t> > *reg) const
{
  /// Decode the bit register, each bit is converted into its own
  /// data-item to make it easier them easier to use (assuming short
  /// ints are 16-bit here).
  /// \param [in] data The input data to be decoded.
  /// \param [in] wordsN The number of words to sort.
  /// \param [in] start The starting index to use for the channel
  /// numbers in the data-items.
  /// \param [out] reg The output register to fill with data-items.
  ///
  for (UInt_t i = 0; i < wordsN * 16; i++)
    (*reg)[i] = (TDataItem<Bool_t>(i + start, (data[0] & (1 << i))));
}

void DataDecoder::DecodeCamac(const UShort_t *data, UInt_t wordsN, UInt_t start,
                           vector<TDataItem<UShort_t> > *module) const
{
  /// Decode CAMAC data.
  /// \param [in] data The input data to be decoded.
  /// \param [in] wordsN The number of words to decode.
  /// \param [in] start The starting index to use for the channel
  /// numbers in the data-items.
  /// \param [out] module The vector of data-items to fill.
  /// \note Words aren't indexed so we add indexing here beginning at
  /// "start"
  ///
  for (UInt_t i = 0; i < wordsN; i++)
    module->push_back(TDataItem<UShort_t>(static_cast<UShort_t>(i + start),
                                          data[i]));
}

void DataDecoder::DecodeFera(const UShort_t *data, UInt_t wordsN, UInt_t start,
                          vector<TDataItem<UShort_t> > *module) const
{
  /// Decode FERA data, depends on VSN as to what we get.  Hard coded
  /// which is wrong but it's faster and less hassle.  For a proper
  /// example, DragonRoot does it properly.
  /// \param [in] data The input data to be decoded.
  /// \param [in] wordsN The number of words to decode.
  /// \param [in] start The starting index to use for the channel
  /// numbers in the data-items.
  /// \param [out] module The vector of data-items to fill.
  /// \note Words should always be with index 0 (from scaler).
  ///
  UInt_t i = 0;
  UShort_t words = data[i++];

  if (words != wordsN - 1 // && words > 0
    ) {
    cerr << "DecodeFera: ERROR - incorrect word count" << endl;
  } else {
    do {
      UShort_t header = data[i++];

      if ((header & 0x8000) == 0x8000) {
        UShort_t vsn = header & 0x00ff;
        switch (vsn) {
        case  1:
        case  2:
        case  3:
        case  4:
        case  5:
        case  6:
        case  7:
        case  8:
        case  9:
        case 10:
        case 11:
        case 12: {
          // silena adc vsn's
          UShort_t vdc = header >> 8 & 0x000f;
          UShort_t pw  = data[i++] & 0x00ff;
          for (UInt_t j = 0; j < vdc && i + j < wordsN; j++) {
            UShort_t sub = (data[i + j] >> 12) & 0x0007;
            UShort_t val = data[i + j] & 0x0fff;
            UShort_t ovf = (data[i + j] >> 15) & 0x0001;
            // printf("vsn %.4x, vdc %.4x, pw %.4x, sub %.4x, val %.4x,"
            //        " ovf %.4x\n", vsn, vdc, pw, sub, val, ovf);
            if (ovf) {
              // cerr << "DecodeFera: WARNING - ADC overflow" << endl;
            } else if (pw & 1 << sub) {
              sub += (vsn - 1) * 8 + start;
              module->push_back(TDataItem<UShort_t>(sub, val));
            } else {
              cerr << "DecodeFera: ERROR - subaddress and pattern-word mismatch"
                   << endl;
            }
          }
          i += vdc;
          break;
        }
        default:
          // unknown vsn's - shouldn't ever get here
          cerr << "DecodeFera: Unknown VSN = " << vsn << endl;
          break;
        }
      }
    } while (i < wordsN);
  }
}

void DataDecoder::DecodeVme(const UInt_t *data, UInt_t wordsN, UInt_t start,
                         vector<TDataItem<UShort_t> > *module) const
{
  /// Decode VME data.  A touch generic but good for the modules that
  /// were used at CRIB.
  /// \param [in] data The input data to be decoded.
  /// \param [in] wordsN The number of words to decode.
  /// \param [in] start The starting index to use for the channel
  /// numbers in the data-items.
  /// \param [out] module The vector of data-items to fill.
  /// \note Lots of debugging info commented out.
  ///
  // V785 ADC is 12-bit
  UInt_t i = 0;
  UInt_t type = 0xffffffff;

  // struct Header_t {
  //   UShort_t geo;
  //   UShort_t type;
  //   UShort_t crate;
  //   UShort_t channels;
  // } header;

  struct Datum_t {
    UShort_t geo;
    UShort_t type;
    UShort_t channel;
    UShort_t valid;
    UShort_t underthreshold;
    UShort_t overflow;
    UShort_t value;
  } datum;

  // struct Eob_t {
  //   UShort_t geo;
  //   UShort_t type;
  //   UShort_t counter;
  // } eob;

  struct Invalid_t {
    UShort_t type;
  } invalid;

  for (i = 0; i < wordsN; i++) {

    type = (data[i] & 0x07000000) >> 24;
    switch (type) {
    case 2:
      // header
      // header.geo      = (data[i] & 0xf8000000) >> 27;
      // header.type     = (data[i] & 0x07000000) >> 24;
      // header.crate    = (data[i] & 0x007f0000) >> 16;
      // header.channels = (data[i] & 0x00003f00) >> 8;
      // cout << "header.geo = "      << header.geo      << endl
      //      << "header.type = "     << header.type     << endl
      //      << "header.crate = "    << header.crate    << endl
      //      << "header.channels = " << header.channels << endl;
      break;
    case 0:
      // valid datum
      datum.geo = (data[i] & 0xf8000000) >> 27;
      // datum.type = (data[i] & 0x07000000) >> 24;
      datum.channel = (data[i] & 0x001f0000) >> 16;
      // datum.valid = (data[i] & 0x00002000) >> 14;
      // datum.underthreshold = (data[i] & 0x00001000) >> 13;
      // datum.overflow = (data[i] & 0x000008000) >> 12;
      datum.value = data[i] & 0x00000fff;
      // cout << "datum.geo = "            << datum.geo            << endl
      //      << "datum.type = "           << datum.type           << endl
      //      << "datum.valid = "          << datum.valid          << endl
      //      << "datum.underthreshold = " << datum.underthreshold << endl
      //      << "datum.overflow = "       << datum.overflow       << endl
      //      << "datum.value = "          << datum.value          << endl;
      module->push_back(TDataItem<UShort_t>((datum.geo - 1) * 32 + 
                                            datum.channel + start,
                                            datum.value));
      break;
    case 4:
      // end-of-block
      // eob.geo     = (data[i] & 0xf8000000) >> 27;
      // eob.type    = (data[i] & 0x07000000) >> 24;
      // eob.counter = data[i] & 0x00ffffff;
      // cout << "eob.geo = "     << eob.geo     << endl
      //      << "eob.type = "    << eob.type    << endl
      //      << "eob.counter = " << eob.counter << endl;
      break;
    case 6:
      // invalid datum
      invalid.type = (data[i] & 0x07000000) >> 24;
      // cerr << "DecodeVme: ERROR - invalid word" << endl;
      break;
    default:
      cerr << "DecodeVme: ERROR - invalid word" << endl;
      break;
    }
  }
}

*/
